#ifndef __JABYENGINE_INTERNAL_CD_TYPES_HPP__
#define __JABYENGINE_INTERNAL_CD_TYPES_HPP__
#include <PSX/AutoLBA/auto_lba.hpp>
#include <PSX/Auxiliary/math_helper.hpp>
#include <PSX/System/IOPorts/cd_io.hpp>
#include <stddef.h>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            enum struct State {
                Free = 0,
                Done = 0,

                Reading,
                BufferFull,
                Error,
            };

            struct FileInfo {
                uint16_t lba;
                uint16_t sectors;

                static constexpr FileInfo from(const AutoLBAEntry& entry) {
                    return FileInfo{entry.lba, CD_IO::DataSector::words_to_sectors(entry.size_words)};
                }
            };

            class SectorBufferAllocator {
            private:
                typedef CD_IO::DataSector* (*AllocatorFunction)(void* ctx);

            private:
                void*             ctx      = nullptr;
                AllocatorFunction allocate = nullptr;

                constexpr SectorBufferAllocator(void* ctx, AllocatorFunction function) : ctx(ctx), allocate(function) {
                }

            public:
                constexpr SectorBufferAllocator() = default;

                static constexpr SectorBufferAllocator create(void* obj, AllocatorFunction function) {
                    return SectorBufferAllocator(obj, function);
                }

                inline CD_IO::DataSector* allocate_sector() const {
                    return this->allocate(this->ctx);
                }
            };

            struct CDTimeStamp {
                static constexpr size_t MaxSector  = 75;
                static constexpr size_t MaxSeconds = 60;

                uint8_t min;
                uint8_t sec;
                uint8_t sector;

                static constexpr CDTimeStamp from(uint16_t lba) {
                    const auto [min, new_lba] = div_and_mod(lba, static_cast<uint16_t>(MaxSector*MaxSeconds));
                    const auto [sec, sectors] = div_and_mod(new_lba, static_cast<uint16_t>(MaxSector));

                    return CDTimeStamp{static_cast<uint8_t>(min), static_cast<uint8_t>(sec), static_cast<uint8_t>(sectors)};
                }

                static constexpr CDTimeStamp from(const FileInfo& file_info) {
                    return CDTimeStamp::from(file_info.lba);
                }

                constexpr uint8_t get_min_cd()    const {
                    return to_bcd(this->min);
                }

                constexpr uint8_t get_sec_cd()    const {
                    return to_bcd(this->sec);
                }

                constexpr uint8_t get_sector_cd() const {
                    return to_bcd(this->sector);
                }
            };
        }
    }
}
#endif //!__JABYENGINE_INTERNAL_CD_TYPES_HPP__